@@ -1,11 +1,16 @@
-/* Copyright (c) 1993-2002
+/* Copyright (c) 2008, 2009
+ *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
+ *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
+ *      Micah Cowan (micah@cowan.name)
+ *      Sadrul Habib Chowdhury (sadrul@users.sourceforge.net)
+ * Copyright (c) 1993-2002, 2003, 2005, 2006, 2007
  *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
  *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
  * Copyright (c) 1987 Oliver Laumann
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
+ * the Free Software Foundation; either version 3, or (at your option)
  * any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -14,20 +19,17 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program (see the file COPYING); if not, write to the
- * Free Software Foundation, Inc.,
- * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * along with this program (see the file COPYING); if not, see
+ * http://www.gnu.org/licenses/, or contact Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
  *
  ****************************************************************
  */
 
-#include "rcs.h"
-RCS_ID("$Id: fileio.c,v 1.10 1994/05/31 12:32:01 mlschroe Exp $ FAU")
-
-
 #include <sys/types.h>
 #include <fcntl.h>
 #include <sys/stat.h>
+#include <pwd.h>
 
 #ifndef SIGINT
 # include <signal.h>
@@ -73,7 +75,7 @@ register char *str1, *str2;
     {
       len2 = strlen(str2);
       if ((cp = realloc(str2, (unsigned) len1 + len2 + add_colon + 1)) == NULL)
-	Panic(0, strnomem);
+	Panic(0, "%s", strnomem);
       bcopy(cp, cp + len1 + add_colon, len2 + 1);
     }
   else
@@ -81,8 +83,8 @@ register char *str1, *str2;
       if (len1 == 0)
 	return 0;
       if ((cp = malloc((unsigned) len1 + add_colon + 1)) == NULL)
-	Panic(0, strnomem);
-      cp[len1 + add_colon] = '\0'; 
+	Panic(0, "%s", strnomem);
+      cp[len1 + add_colon] = '\0';
     }
   bcopy(str1, cp, len1);
   if (add_colon)
@@ -98,6 +100,42 @@ char *rcfile;
   char buf[256];
   char *p;
 
+  /* Tilde prefix support courtesy <hesso@pool.math.tu-berlin.de>,
+   * taken from a Debian patch. */
+  if (rcfile && *rcfile == '~')
+    {
+      static char rcfilename_tilde_exp[MAXPATHLEN+1];
+      char *slash_position = strchr(rcfile, '/');
+      if (slash_position == rcfile+1)
+        {
+          char *home = getenv("HOME");
+          if (!home)
+            {
+              Msg(0, "%s: source: tilde expansion failed", rc_name);
+              return NULL;
+            }
+          snprintf(rcfilename_tilde_exp, MAXPATHLEN, "%s/%s", home, rcfile+2);
+        }
+      else if (slash_position)
+        {
+          struct passwd *p;
+          *slash_position = 0;
+          p = getpwnam(rcfile+1);
+          if (!p)
+            {
+              Msg(0, "%s: source: tilde expansion failed for user %s", rc_name, rcfile+1);
+              return NULL;
+            }
+          snprintf(rcfilename_tilde_exp, MAXPATHLEN, "%s/%s", p->pw_dir, slash_position+1);
+        }
+      else
+        {
+          Msg(0, "%s: source: illegal tilde expression.", rc_name);
+          return NULL;
+        }
+      rcfile = rcfilename_tilde_exp;
+    }
+
   if (rcfile)
     {
       char *rcend = rindex(rc_name, '/');
@@ -132,14 +170,16 @@ char *rcfile;
  * 1) rcfilename = "/etc/screenrc"
  * 2) rcfilename = RcFileName
  */
-void
-StartRc(rcfilename)
+int
+StartRc(rcfilename, nopanic)
 char *rcfilename;
+int nopanic;
 {
   register int argc, len;
   register char *p, *cp;
   char buf[2048];
   char *args[MAXARGS];
+  int argl[MAXARGS];
   FILE *fp;
   char *oldrc_name = rc_name;
 
@@ -152,29 +192,31 @@ char *rcfilename;
 
   rc_name = findrcfile(rcfilename);
 
-  if ((fp = secfopen(rc_name, "r")) == NULL)
+  if (rc_name == NULL || (fp = secfopen(rc_name, "r")) == NULL)
     {
-      if (!rc_recursion && RcFileName && !strcmp(RcFileName, rc_name))
+      const char *rc_nonnull = rc_name ? rc_name : rcfilename;
+      if (!rc_recursion && RcFileName && !strcmp(RcFileName, rc_nonnull))
 	{
           /*
            * User explicitly gave us that name,
            * this is the only case, where we get angry, if we can't read
            * the file.
            */
-	  debug3("StartRc: '%s','%s', '%s'\n", RcFileName, rc_name, rcfilename);
-          Panic(0, "Unable to open \"%s\".", rc_name);
-	  /* NOTREACHED */
+	  debug3("StartRc: '%s','%s', '%s'\n", RcFileName, rc_name ? rc_name : "(null)", rcfilename);
+          if (!nopanic) Panic(0, "Unable to open \"%s\".", rc_nonnull);
+	  /* possibly NOTREACHED */
 	}
-      debug1("StartRc: '%s' no good. ignored\n", rc_name);
-      Free(rc_name);
+      debug1("StartRc: '%s' no good. ignored\n", rc_nonnull);
+      if (rc_name)
+	Free(rc_name);
       rc_name = oldrc_name;
-      return;
+      return 1;
     }
   while (fgets(buf, sizeof buf, fp) != NULL)
     {
       if ((p = rindex(buf, '\n')) != NULL)
 	*p = '\0';
-      if ((argc = Parse(expand_vars(buf, display), args)) == 0)
+      if ((argc = Parse(buf, sizeof buf, args, argl)) == 0)
 	continue;
       if (strcmp(args[0], "echo") == 0)
 	{
@@ -189,7 +231,7 @@ char *rcfilename;
 	  if (argc != 3)
 	    {
 	      AddStr("\r\n");
-	      Flush();
+	      Flush(0);
 	    }
 	}
       else if (strcmp(args[0], "sleep") == 0)
@@ -202,7 +244,7 @@ char *rcfilename;
 	      Msg(0, "%s: sleep: one numeric argument expected.", rc_name);
 	      continue;
 	    }
-	  DisplaySleep(atoi(args[1]), 1);
+	  DisplaySleep1000(1000 * atoi(args[1]), 1);
 	}
 #ifdef TERMINFO
       else if (!strcmp(args[0], "termcapinfo") || !strcmp(args[0], "terminfo"))
@@ -241,7 +283,7 @@ char *rcfilename;
 	  if (rc_recursion <= 10)
 	    {
 	      rc_recursion++;
-	      StartRc(args[1]);
+	      (void)StartRc(args[1], 0);
 	      rc_recursion--;
 	    }
 	}
@@ -249,6 +291,7 @@ char *rcfilename;
   fclose(fp);
   Free(rc_name);
   rc_name = oldrc_name;
+  return 0;
 }
 
 void
@@ -261,30 +304,32 @@ char *rcfilename;
 
   rc_name = findrcfile(rcfilename);
 
-  if ((fp = secfopen(rc_name, "r")) == NULL)
+  if (rc_name == NULL || (fp = secfopen(rc_name, "r")) == NULL)
     {
+      const char *rc_nonnull = rc_name ? rc_name : rcfilename;
       if (rc_recursion)
-        Msg(errno, "%s: source %s", oldrc_name, rc_name);
-      else if (RcFileName && !strcmp(RcFileName, rc_name))
+        Msg(errno, "%s: source %s", oldrc_name, rc_nonnull);
+      else if (RcFileName && !strcmp(RcFileName, rc_nonnull))
 	{
     	  /*
  	   * User explicitly gave us that name, 
 	   * this is the only case, where we get angry, if we can't read
 	   * the file.
 	   */
-  	  debug3("FinishRc:'%s','%s','%s'\n", RcFileName, rc_name, rcfilename);
-          Panic(0, "Unable to open \"%s\".", rc_name);
+  	  debug3("FinishRc:'%s','%s','%s'\n", RcFileName, rc_name ? rc_name : "(null)", rcfilename);
+          Panic(0, "Unable to open \"%s\".", rc_nonnull);
 	  /* NOTREACHED */
 	}
-      debug1("FinishRc: '%s' no good. ignored\n", rc_name);
-      Free(rc_name);
+      debug1("FinishRc: '%s' no good. ignored\n", rc_nonnull);
+      if (rc_name)
+	Free(rc_name);
       rc_name = oldrc_name;
       return;
     }
 
   debug("finishrc is going...\n");
   while (fgets(buf, sizeof buf, fp) != NULL)
-    RcLine(buf);
+    RcLine(buf, sizeof buf);
   (void) fclose(fp);
   Free(rc_name);
   rc_name = oldrc_name;
@@ -311,10 +356,12 @@ char *rcfilename;
  * This is bad when we run detached.
  */
 void
-RcLine(ubuf)
+RcLine(ubuf, ubufl)
 char *ubuf;
+int ubufl;
 {
-  char *args[MAXARGS], *buf;
+  char *args[MAXARGS];
+  int argl[MAXARGS];
 #ifdef MULTIUSER
   extern struct acluser *EffectiveAclUser;	/* acl.c */
   extern struct acluser *users;		/* acl.c */
@@ -327,19 +374,17 @@ char *ubuf;
     }
   else
     flayer = fore ? fore->w_savelayer : 0;
-  buf = expand_vars(ubuf, display);
-  if (Parse(buf, args) <= 0)
+  if (Parse(ubuf, ubufl, args, argl) <= 0)
     return;
 #ifdef MULTIUSER
   if (!display)
     {
       /* the session owner does it, when there is no display here */
       EffectiveAclUser = users;        
-      debug1("RcLine: WARNING, no display no user! Session owner does: %s",
-             ubuf);
+      debug("RcLine: WARNING, no display no user! Session owner executes command\n");
     }
 #endif
-  DoCommand(args);
+  DoCommand(args, argl);
 #ifdef MULTIUSER
   EffectiveAclUser = 0;
 #endif
@@ -385,7 +430,7 @@ int dump;
       if (fn == 0)
 	{
 	  i = SockName - SockPath;
-	  if (i > sizeof(fnbuf) - 9)
+	  if (i > (int)sizeof(fnbuf) - 9)
 	    i = 0;
 	  strncpy(fnbuf, SockPath, i);
 	  strcpy(fnbuf + i, ".termcap");
@@ -483,6 +528,7 @@ int dump;
 		}
 	      if (dump == DUMP_SCROLLBACK)
 		{
+#ifdef COPY_PASTE
 		  for (i = 0; i < fore->w_histheight; i++)
 		    {
 		      p = (char *)(WIN(i)->image);
@@ -492,6 +538,7 @@ int dump;
 			putc(p[j], f);
 		      putc('\n', f);
 		    }
+#endif
 		}
 	      for (i = 0; i < fore->w_height; i++)
 		{
@@ -580,7 +627,7 @@ int *lenp;
   if ((buf = malloc(size)) == NULL)
     {
       close(i);
-      Msg(0, strnomem);
+      Msg(0, "%s", strnomem);
       return NULL;
     }
   errno = 0;
@@ -782,7 +829,7 @@ char *cmd;
 #ifdef SIGPIPE
       signal(SIGPIPE, SIG_DFL);
 #endif
-      execl("/bin/sh", "sh", "-c", cmd, 0);
+      execl("/bin/sh", "sh", "-c", cmd, (char *)0);
       Panic(errno, "/bin/sh");
     default:
       break;
@@ -796,6 +843,7 @@ readpipe(cmdv)
 char **cmdv;
 {
   int pi[2];
+
   if (pipe(pi))
     {
       Msg(errno, "pipe");
@@ -814,15 +862,22 @@ char **cmdv;
 #endif
       close(1);
       if (dup(pi[1]) != 1)
-        Panic(0, "dup");
+	{
+	  close(pi[1]);
+          Panic(0, "dup");
+	}
       closeallfiles(1);
       if (setgid(real_gid) || setuid(real_uid))
-        Panic(errno, "setuid/setgid");
+	{
+          close(1);
+          Panic(errno, "setuid/setgid");
+	}
 #ifdef SIGPIPE
       signal(SIGPIPE, SIG_DFL);
 #endif
       execvp(*cmdv, cmdv);
-      Panic(errno, *cmdv);
+      close(1);
+      Panic(errno, "%s", *cmdv);
     default:
       break;
     }
